"""
Module for uniform treatment of dataclasses and attrs classes.
"""
import dataclasses
from abc import abstractmethod
from dataclasses import dataclass
from typing import TypeVar, Generic, Mapping, Type, Union

from datargs.meta import AbstractClassProperty

FieldType = TypeVar("FieldType")


@dataclass
class DatargsParams:
    parser: dict = dataclasses.field(default_factory=dict)
    sub_commands: dict = dataclasses.field(default_factory=dict)
    name: str = None

    def __post_init__(self, *args, **kwargs):
        for key, value in (
            ("required", True),
            ("dest", "__datargs_dest__"),
        ):
            self.sub_commands.setdefault(key, value)


class RecordField(Generic[FieldType]):
    """
    Abstract base class for fields of dataclasses or attrs classes.
    """

    _field: FieldType
    NONE = object()

    def __init__(self, field):
        self._field = field

    @abstractmethod
    def is_required(self) -> bool:
        """
        Return whether field is required.
        """
        pass

    def has_default(self) -> bool:
        """
        Helper method to indicate whether a field has a default value.
        Used to make intention clearer in call sites.
        """
        return not self.is_required()

    @property
    def default(self):
        return self._field.default

    @property
    def name(self):
        return self._field.name

    @property
    def type(self):
        return self._field.type

    @property
    def metadata(self):
        return self._field.metadata

    @property
    def is_positional(self) -> bool:
        return self.metadata.get("positional", False)


class DataField(RecordField[dataclasses.Field]):
    """
    Represents a dataclass field.
    """

    def is_required(self) -> bool:
        return self.default is dataclasses.MISSING


class NotARecordClass(Exception):
    pass


class RecordClass(Generic[FieldType]):
    """
    Abstract base class for dataclasses or attrs classes.
    """

    # The name of the attribute that holds field definitions
    fields_attribute: str = AbstractClassProperty()

    # The type to wrap fields with
    field_wrapper_type: Type[RecordField] = AbstractClassProperty()
    _implementors = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        if not getattr(cls, "__abstractmethods__", None):
            cls._implementors.append(cls)

    def __init__(self, cls):
        self.cls: type = cls

    @property
    def datargs_params(self) -> DatargsParams:
        return getattr(self.cls, "__datargs_params__", DatargsParams())

    @property
    def parser_params(self):
        return self.datargs_params.parser

    @property
    def sub_commands_params(self):
        return self.datargs_params.sub_commands

    @property
    def name(self):
        return self.cls.__name__

    @abstractmethod
    def fields_dict(self) -> Mapping[str, FieldType]:
        """
        Returns a mapping of field names to field wrapper classes.
        """
        pass

    @classmethod
    def can_wrap_class(cls, potential_record_class) -> bool:
        """
        Returns whether this class is the appropriate implementation for wrapping `potential_record_class`.
        """
        return getattr(potential_record_class, cls.fields_attribute, None) is not None

    @classmethod
    def wrap_class(cls, record_class) -> "RecordClass":
        """
        Wrap `record_class` with the appropriate wrapper.
        """
        for candidate in cls._implementors:
            if candidate.can_wrap_class(record_class):
                return candidate(record_class)
        if getattr(record_class, "__attrs_attrs__", None) is not None:
            raise NotARecordClass(
                f"can't accept '{record_class.__name__}' because it is an attrs class and attrs is not installed"
            )
        raise NotARecordClass(
            f"class '{record_class.__name__}' is not a dataclass nor an attrs class"
        )

    @classmethod
    def get_field(cls, field: FieldType):
        """
        Wrap field with field classes with a uniform interface.
        """
        return cls.field_wrapper_type(field)


class DataClass(RecordClass[dataclasses.Field]):
    """
    Represents a dataclass.
    """

    fields_attribute = "__dataclass_fields__"
    field_wrapper_type = DataField

    def fields_dict(self) -> Mapping[str, FieldType]:
        fields = dataclasses.fields(self.cls)
        return {field.name: self.get_field(field) for field in fields}


try:
    import attr
except ImportError:
    pass
else:
    # import class to register it in RecordClass._implementors
    import datargs.compat.attrs


def is_optional(typ):
    return (
        getattr(typ, "__origin__", None) is Union
        and len(typ.__args__) == 2
        and type(None) in typ.__args__
    )



# noinspection PyUnresolvedReferences
"""
Declarative, type safe `argparse` parsers.

>>> @dataclass
... class Args:
...     just_a_string: str
...     num: int
...     store_true: bool = False
...     store_false: bool = True
>>> args = parse(Args, ["--just-a-string", "STRING", "--num", "0", "--store-true", "--store-false"])
>>> args
Args(just_a_string='STRING', num=0, store_true=True, store_false=False)

Pycharm correctly infers that `args` is of type `Args`.
Trying to access a non-existent member is a type error:
>>> args.nope  # doctest: +SKIP
Pycharm says: Unresolved attribute reference 'nope' for class 'Args'

A flag with no defaults is assumed to be False by default:
>>> @dataclass
... class Args:
...     no_default: bool
>>> parse(Args, [])
Args(no_default=False)

Enums are supported. They should be specified by name on the command line:
>>> class FoodEnum(Enum):
...     gnocchi = 0
...     kimchi = 1
>>> @dataclass
... class Args:
...     food: FoodEnum
>>> parse(Args, ["--food", "kimchi"])
Args(food=<FoodEnum.kimchi: 1>)
>>> parse(Args, ["--food", "poutine"]) # doctest: +SKIP
usage: make.py [-h] --food {gnocchi,kimchi}
make.py: error: argument --food: 'poutine': invalid value
...
SystemExit: 2

Specifying enums by name is not currently supported.
"""
import dataclasses

# noinspection PyUnresolvedReferences,PyProtectedMember
from argparse import (
    ArgumentParser,
    ArgumentTypeError,
    _SubParsersAction,
    Namespace,
    _CountAction,
)
from dataclasses import dataclass, MISSING
from enum import Enum
from functools import wraps, partial
from inspect import signature
from typing import (
    Callable,
    Dict,
    TypeVar,
    Type,
    Sequence,
    Optional,
    overload,
    Any,
    Union,
    cast,
    get_type_hints,
    List,
    get_origin,
    Literal,
)

from boltons.strutils import camel2under

from .compat import (
    RecordField,
    RecordClass,
    NotARecordClass,
    DatargsParams,
    is_optional,
)


@dataclass
class Action:
    args: Sequence[Any] = dataclasses.field(default_factory=list)
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)


DispatchCallback = Callable[[str, RecordField, dict], Action]
AddArgFunc = Callable[[RecordField, dict], Action]


def field_name_to_arg_name(name: str, positional=False) -> str:
    if positional:
        return name
    return f"--{name.replace('_','-')}"


SpecialRule = Callable[[Type["TypeDispatch"], RecordField], Optional[Action]]


def is_subclass(cls, parent):
    try:
        return cls is parent or issubclass(cls, parent)
    except TypeError:
        return False


class TypeDispatch:

    dispatch: Dict[type, AddArgFunc] = {}
    special_rules: List[SpecialRule] = []

    @classmethod
    def add_arg(cls, field: RecordField, override: dict):
        typ = override.get("type", field.type)
        override = {**override, "type": typ}
        dispatch_type = get_origin(typ) or typ
        for typ, func in cls.dispatch.items():
            if is_subclass(dispatch_type, typ):
                return func(field, override)
        return add_any(field, override)

    @classmethod
    def register(cls, typ):
        def decorator(func: DispatchCallback) -> AddArgFunc:
            cls.dispatch[typ] = new_func = wraps(func)(add_name_formatting(func))
            return new_func

        return decorator


def add_name_formatting(func: DispatchCallback) -> AddArgFunc:
    @wraps(func)
    def new_func(field: RecordField, override: dict):
        return func(
            field_name_to_arg_name(field.name, positional=field.is_positional),
            field,
            override,
        )

    return new_func


@add_name_formatting
def add_any(name: str, field: RecordField, extra: dict) -> Action:
    return add_default(name, field, extra)


def get_option_strings(name: str, field: RecordField):
    return [name, *field.metadata.get("aliases", [])]


def common_kwargs(field: RecordField):
    return {"type": field.type, **subdict(field.metadata, ["aliases", "positional"])}


def subdict(dct, remove_keys):
    return {key: value for key, value in dct.items() if key not in remove_keys}


def add_default(name, field: RecordField, override: dict) -> Action:
    override = {
        "default": field.default,
        **common_kwargs(field),
        **override,
    }
    if not field.is_positional and (override.get("default", object()) is not None):
        override["required"] = field.is_required()
    return Action(kwargs=override, args=get_option_strings(name, field))


T = TypeVar("T")


def call_func_with_matching_kwargs(func: Callable[..., T], *args, **kwargs) -> T:
    sig = signature(func)
    new_kwargs = {key: value for key, value in kwargs.items() if key in sig.parameters}
    return func(*args, **new_kwargs)


@TypeDispatch.register(str)
def add_str(name, field, override: dict):
    return add_default(name, field, override)


@TypeDispatch.register(Sequence)
def sequence_arg(name: str, field: RecordField, override: dict) -> Action:
    nargs = field.metadata.get("nargs")
    if nargs:
        assert nargs in ("+", "?") or isinstance(nargs, int)
    else:
        nargs = "+" if field.is_required() else "*"
    new_type = (override.get("type") or field.type).__args__[0]
    return TypeDispatch.add_arg(
        field,
        {**override, "nargs": nargs, "type": new_type},
    )


@TypeDispatch.register(Union)
def union_arg(name: str, field: RecordField, override: dict) -> Action:
    inner_type = (set(field.type.__args__) - {type(None)}).pop()
    return TypeDispatch.add_arg(
        field, {**override, "default": None, "type": inner_type}
    )


@TypeDispatch.register(Literal)
def literal_arg(name: str, field: RecordField, override: dict) -> Action:
    typ = override.get("type", field.type)
    literal_options = set(typ.__args__)

    if "choices" in field.metadata:
        choices = set(field.metadata["choices"])
        assert choices.issubset(literal_options)
    else:
        choices = literal_options

    assert (
        len(set(map(type, choices))) == 1
    ), "All choices in literal must have the same type!"
    inner_type = type(typ.__args__[0])

    return TypeDispatch.add_arg(
        field, {**override, "choices": choices, "type": inner_type}
    )


@TypeDispatch.register(bool)
def bool_arg(name: str, field: RecordField, override: dict) -> Action:
    kwargs = {
        **subdict(common_kwargs(field), ["type"]),
        **override,
        "action": "store_false"
        if field.default and field.has_default()
        else "store_true",
    }
    kwargs.pop("type", None)
    return Action(
        args=get_option_strings(name, field),
        kwargs=kwargs,
    )


@TypeDispatch.register(Enum)
def enum_arg(name: str, field: RecordField, override: dict) -> Action:
    field_type = override.get("type") or field.type

    def enum_type_func(value: str):
        result = field_type.__members__.get(value)
        if not result:
            raise ArgumentTypeError(
                f"invalid choice: {value!r} (choose from {[e.name for e in field.type]})"
            )
        return result

    return add_default(
        name,
        field,
        {
            **override,
            "type": enum_type_func,
            "choices": field_type,
            "metavar": f"{{{','.join(field_type.__members__)}}}",
        },
    )


ParserType = TypeVar("ParserType", bound=ArgumentParser)


@overload
def make_parser(cls: type) -> ArgumentParser:
    pass


@overload
def make_parser(cls: type, parser: None = None) -> ArgumentParser:
    pass


@overload
def make_parser(cls: type, parser: ParserType) -> ParserType:
    pass


def make_parser(cls, parser=None):
    # noinspection PyShadowingNames
    """
    Create parser that parses command-line arguments according to the fields of `cls`.
    Use this if you want to do anything with the parser other than immediately parsing the command-line arguments.
    If you do want to parse immediately, use `parse()`.
    :param cls: class according to which argument parser is created
    :param parser: parser to add arguments to, by default creates a new parser
    :return: instance of `parser_cls` which parses command line according to `cls`

    >>> @dataclass
    ... class Args:
    ...     first_arg: int
    >>> parse(Args, ["--first-arg", "0"])
    Args(first_arg=0)
    >>> parser = make_parser(Args)
    >>> parser.add_argument("--second-arg", type=float) # doctest: +ELLIPSIS
    [...]
    >>> parser.parse_args(["--first-arg", "0", "--second-arg", "1.5"])
    Namespace(first_arg=0, second_arg=1.5)
    """
    record_class = RecordClass.wrap_class(cls)
    return _make_parser(record_class, parser=parser)


class DatargsSubparsers(_SubParsersAction):
    """
    A subparsers action that creates the correct sub-command class upon parsing.
    """

    def __init__(self, name, *args, **kwargs):
        self.__name = name
        super().__init__(*args, **kwargs)
        self._command_type_map = {}

    def add_parser(self, typ: type, name: str, aliases=(), *args, **kwargs):
        result = super().add_parser(name, aliases=aliases, *args, **kwargs)
        for alias in [name, *aliases]:
            self._command_type_map[alias] = typ
        return result

    def __call__(self, parser, namespace, values, *args, **kwargs):
        new_ns = Namespace()
        name, *_ = values
        super().__call__(parser, new_ns, values)

        if hasattr(new_ns, self.dest):
            delattr(new_ns, self.dest)

        setattr(namespace, self.__name, self._command_type_map[name](**vars(new_ns)))


def _make_parser(record_class: RecordClass, parser: ParserType = None) -> ParserType:
    if not parser:
        parser = ArgumentParser(**record_class.parser_params)
    assert parser is not None
    for name, field in record_class.fields_dict().items():
        sub_commands = None
        try:
            if field.type.__origin__ is Union and not is_optional(field.type):
                sub_commands = field.type.__args__
        except AttributeError:
            pass
        if sub_commands is not None:
            add_subparsers(parser, record_class, field, sub_commands)
        else:
            action = TypeDispatch.add_arg(field, {})
            fix_count_action_kwargs(action)
            parser.add_argument(*action.args, **action.kwargs)
    return parser


def fix_count_action_kwargs(action):
    """
    If argument action is "count", remove "type" kwarg
    """
    if action.kwargs.get("action") == "count":
        action.kwargs.pop("type")


def add_subparsers(
    parser: ArgumentParser,
    top_class: RecordClass,
    sub_parsers_field: RecordField,
    sub_parser_classes: Sequence[type],
):
    # noinspection PyArgumentList

    subparsers = cast(
        DatargsSubparsers,
        parser.add_subparsers(
            **top_class.sub_commands_params,
            action=DatargsSubparsers,
            name=sub_parsers_field.name,
        ),
    )
    for command in sub_parser_classes:
        try:
            sub_parsers_args = top_class.wrap_class(command)
        except NotARecordClass:
            raise Exception(
                f"{top_class.name}.{sub_parsers_field.name}: "
                f"Union must be used with dataclass/attrs class and creates a subparser (got: {sub_parsers_field.type})"
            )
        sub_parser = subparsers.add_parser(
            command,
            sub_parsers_args.datargs_params.name
            or camel2under(sub_parsers_args.name).replace("_", "-"),
            **sub_parsers_args.parser_params,
        )
        _make_parser(sub_parsers_args, sub_parser)


def parse(cls: Type[T], args: Optional[Sequence[str]] = None, *, parser=None) -> T:
    """
    Parse command line arguments according to the fields of `cls` and populate it.
    Accepts classes decorated with `dataclass` or `attr.s`.
    :param cls: class to parse command-line arguments by
    :param parser: existing parser to add arguments to and parse from
    :param args: arguments to parse (default: `sys.arg`)
    :return: an instance of cls

    >>> @dataclass
    ... class Args:
    ...     is_flag: bool
    ...     num: int = 0
    >>> parse(Args, ["--num", "1"])
    Args(is_flag=False, num=1)
    """
    result = vars(make_parser(cls, parser=parser).parse_args(args))
    try:
        command_dest = cls.__datargs_params__.sub_commands.get("dest", None)
    except AttributeError:
        pass
    else:
        if command_dest is not None and command_dest in result:
            del result[command_dest]
    return cls(**result)


def argsclass(
    cls: type = None,
    *args,
    description: str = None,
    parser_params: dict = None,
    name: str = None,
    **kwargs,
):
    """
    A wrapper around `dataclass` for passing `description` and other params (in `parser_params`)
    to the `ArgumentParser` constructor.

    :param cls: class to wrap
    :param description: parser description
    :param parser_params: dict of arguments to parser class constructor
    :param name: Only used in subcommands, string to invoke subcommand. By default, the string used is the class'
        name converted to kebab-case, i.e., snake case with hyphens
    """
    # sub_commands_params has been disabled until a useful use case is found
    datargs_kwargs = {
        "description": description,
        "parser_params": parser_params,
        "name": name,
        "sub_commands_params": {},
    }

    if cls is None:
        # We're called with parens.
        return partial(
            make_class,
            *args,
            **datargs_kwargs,
            **kwargs,
        )

    return make_class(
        cls,
        *args,
        **datargs_kwargs,
        **kwargs,
    )


def make_class(
    cls,
    description: str = None,
    parser_params: dict = None,
    name: str = None,
    sub_commands_params: dict = None,
    *args,
    **kwargs,
):
    try:
        RecordClass.wrap_class(cls)
    except NotARecordClass:
        for key, value in cls.__dict__.items():
            if not isinstance(value, dataclasses.Field) or value.default is not MISSING:
                continue
            typ = value.type or get_type_hints(cls)[key]
            if typ is bool:
                value.default = False
        new_cls = dataclass(*args, **kwargs)(cls)
    else:
        new_cls = cls
    new_cls.__datargs_params__ = DatargsParams(
        parser={"description": description, **(parser_params or {})},
        sub_commands=sub_commands_params or {},
        name=name,
    )
    return new_cls


def arg(
    positional=False,
    nargs=None,
    const=None,
    default=MISSING,
    choices=None,
    help=None,
    metavar=None,
    aliases: Sequence[str] = (),
    **kwargs,
):
    """
    Helper method to more easily add parsing-related behavior.
    Supports aliases:
    >>> @dataclass
    ... class Args:
    ...     num: int = arg(aliases=["-n"])
    >>> parse(Args, ["--num", "0"])
    Args(num=0)
    >>> parse(Args, ["-n", "0"])
    Args(num=0)

    Accepts all arguments to both `ArgumentParser.add_argument` and `dataclass.field`:
    >>> @dataclass
    ... class Args:
    ...     invisible_arg: int = arg(default=0, repr=False, metavar="MY_ARG", help="argument description")
    >>> print(Args())
    Args()
    >>> make_parser(Args).print_help() # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    usage: ...
      --invisible-arg MY_ARG    argument description
    """
    return dataclasses.field(
        metadata=remove_dict_nones(
            dict(
                nargs=nargs,
                choices=choices,
                const=const,
                help=help,
                metavar=metavar,
                aliases=aliases,
                positional=positional,
            )
        ),
        default=default,
        **kwargs,
    )


def remove_dict_nones(dct: dict) -> dict:
    return {key: value for key, value in dct.items() if value is not None}


if __name__ == "__main__":
    import doctest

    OC = doctest.OutputChecker

    class AEOutputChecker(OC):
        def check_output(self, want, got, optionflags):
            if optionflags & doctest.ELLIPSIS:
                want = want.replace("[...]", doctest.ELLIPSIS_MARKER)
            return super().check_output(want, got, optionflags)

    doctest.OutputChecker = AEOutputChecker
    doctest.testmod(optionflags=doctest.REPORT_NDIFF)



from argparse import ArgumentParser
from textwrap import indent, dedent
from typing import Tuple


def convert(parser: ArgumentParser):
    print(convert_str(parser))


def convert_str(parser: ArgumentParser) -> str:
    attrs, is_argsclass = get_attrs(parser)
    if is_argsclass:
        header = dedent(
            """
        from datargs import argsclass, arg
        
        
        @argsclass
        class Args:
        """
        )
    else:
        header = dedent(
            f"""
        from dataclasses import dataclass


        @dataclass
        class Args:
        """
        )

    return header + indent(attrs, " " * 4)


def get_attrs(parser: ArgumentParser) -> Tuple[str, bool]:
    def to_var(x):
        return x.strip("-").replace("-", "_")

    seen = set()
    is_argsclass = False
    result = ""
    for action in parser._actions:
        first, *aliases = action.option_strings
        if first in seen or first == "-h":
            continue
        seen.add(first)
        name = to_var(first)
        aliases = [f"aliases={aliases}"] if aliases else []
        try:
            typ = action.type.__name__
        except AttributeError:
            if action.type is not None:
                raise
            typ = action.type or "str"
        relevant_pairs = {
            key: value
            for key, value in vars(action).items()
            if value is not None
            and key
            not in [
                "required",
                "dest",
                "const",
                "option_strings",
                "container",
                "nargs",
                "type",
            ]
        }
        rest = [f"{key}={value!r}" for key, value in relevant_pairs.items()]
        result += f"{name}: {typ}"
        if set(relevant_pairs) - {"default"}:
            is_argsclass = True
            result += f" = arg({','.join(aliases + rest)})"
        elif not action.required:
            default = vars(action).get("default", None)
            result += f" = {default!r}"
        result += "\n"
    return result, is_argsclass


def random_test():
    parser = ArgumentParser()
    parser.add_argument("-a", required=True)
    parser.add_argument("-b", default=3, type=int)
    print(convert(parser))
    parser = ArgumentParser()
    parser.add_argument("-a", required=True, help="help")
    parser.add_argument("-b", default=3, type=int)
    print(convert(parser))
    parser = ArgumentParser()
    parser.add_argument("-a", type=str, required=True, help="help")
    parser.add_argument("-b", default=3, type=int, help="help")
    print(convert(parser))


if __name__ == "__main__":
    main()



from . import RecordField, RecordClass

try:
    import attr
except ImportError:
    pass
else:

    class AttrField(RecordField[attr.Attribute]):
        def is_required(self) -> bool:
            return self.default is attr.NOTHING

    class AttrClass(RecordClass[attr.Attribute]):
        fields_attribute = "__attrs_attrs__"
        field_wrapper_type = AttrField

        def fields_dict(self):
            return {
                name: self.get_field(field)
                for name, field in attr.fields_dict(self.cls).items()
            }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datargs', 'datargs.compat']

package_data = \
{'': ['*']}

install_requires = \
['boltons>=20.2.1,<21.0.0']

extras_require = \
{':python_version == "3.6"': ['dataclasses>=0.6,<0.7'],
 'attrs': ['attrs>=20.2.0,<21.0.0']}

setup_kwargs = {
    'name': 'datargs',
    'version': '0.10.0',
    'description': 'Declarative, type-safe command line argument parsers from dataclasses and attrs classes',
    'long_description': '# datargs\n\nA paper-thin wrapper around `argparse` that creates type-safe parsers\nfrom `dataclass` and `attrs` classes.\n\n## Quickstart\n\n\nInstall `datargs`:\n\n```bash\npip install datargs\n```\n\nCreate a `dataclass` (or an `attrs` class) describing your command line interface, and call\n`datargs.parse()` with the class:\n\n```python\n# script.py\nfrom dataclasses import dataclass\nfrom pathlib import Path\nfrom datargs import parse\n\n@dataclass  # or @attr.s(auto_attribs=True)\nclass Args:\n    url: str\n    output_path: Path\n    verbose: bool\n    retries: int = 3\n\ndef main():\n    args = parse(Args)\n    print(args)\n\nif __name__ == "__main__":\n    main()\n```\n\n***(experimental)*** Alternatively: convert an existing parser to a dataclass:\n```python\n# script.py\nparser = ArgumentParser()\nparser.add_argument(...)\nfrom datargs import convert\nconvert(parser)\n```\n\n`convert()` prints a class definition to the console.\nCopy it to your script.\n\nMypy and pycharm correctly infer the type of `args` as `Args`, and your script is good to go!\n```bash\n$ python script.py -h\nusage: test.py [-h] --url URL --output-path OUTPUT_PATH [--retries RETRIES]\n               [--verbose]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --url URL\n  --output-path OUTPUT_PATH\n  --retries RETRIES\n  --verbose\n$ python script.py --url "https://..." --output-path out --retries 4 --verbose\nArgs(url="https://...", output_path=Path("out"), retries=4, verbose=True)\n```\n\n## Table of Contents\n\n<!-- toc -->\n\n- [Features](#features)\n  * [Static verification](#static-verification)\n  * [`dataclass`/`attr.s` agnostic](#dataclassattrs-agnostic)\n  * [Aliases](#aliases)\n  * [`ArgumentParser` options](#argumentparser-options)\n  * [Enums](#enums)\n  * [Sequences, Optionals, and Literals](#sequences-optionals-and-literals)\n  * [Sub Commands](#sub-commands)\n- ["Why not"s and design choices](#why-nots-and-design-choices)\n  * [Just use argparse?](#just-use-argparse)\n  * [Use `click`](#use-clickhttpsclickpalletsprojectscomen7x)?\n  * [Use `clout`](#use-clouthttpscloutreadthedocsioenlatestindexhtml)?\n  * [Use `simple-parsing`](#use-simple-parsinghttpspypiorgprojectsimple-parsing)?\n  * [Use `argparse-dataclass`](#use-argparse-dataclasshttpspypiorgprojectargparse-dataclass)?\n  * [Use `argparse-dataclasses`](#use-argparse-dataclasseshttpspypiorgprojectargparse-dataclasses)?\n- [FAQs](#faqs)\n  * [Is this cross-platform?](#is-this-cross-platform)\n  * [Why are mutually exclusive options not supported?](#why-are-mutually-exclusive-options-not-supported)\n\n<!-- tocstop -->\n\n## Features\n\n### Static verification\nMypy/Pycharm have your back when you when you make a mistake:\n```python\n...\ndef main():\n    args = parse(Args)\n    args.urll  # typo\n...\n```\nPycharm says: `Unresolved attribute reference \'urll\' for class \'Args\'`.\n\nMypy says: `script.py:15: error: "Args" has no attribute "urll"; maybe "url"?`\n\n\n### `dataclass`/`attr.s` agnostic\n```pycon\n>>> import attr, datargs\n>>> @attr.s\n... class Args:\n...     flag: bool = attr.ib()\n>>> datargs.parse(Args, [])\nArgs(flag=False)\n```\n\n### Aliases\nAliases and `ArgumentParser.add_argument()` parameters are taken from `metadata`:\n\n```pycon\n>>> from dataclasses import dataclass, field\n>>> from datargs import parse\n>>> @dataclass\n... class Args:\n...     retries: int = field(default=3, metadata=dict(help="number of retries", aliases=["-r"], metavar="RETRIES"))\n>>> parse(Args, ["-h"])\nusage: ...\noptional arguments:\n  -h, --help            show this help message and exit\n  --retries RETRIES, -r RETRIES\n>>> parse(Args, ["-r", "4"])\nArgs(retries=4)\n```\n\n`arg` is a replacement for `field` that puts `add_argument()` parameters in `metadata`.\nUse it to save precious keystrokes:\n```pycon\n>>> from dataclasses import dataclass\n>>> from datargs import parse, arg\n>>> @dataclass\n... class Args:\n...     retries: int = arg(default=3, help="number of retries", aliases=["-r"], metavar="RETRIES")\n>>> parse(Args, ["-h"])\n# exactly the same as before\n```\n\n**NOTE**: `arg()` does not currently work with `attr.s`.\n\n`arg()` also supports all `field`/`attr.ib()` keyword arguments.\n\n\n### `ArgumentParser` options\nYou can pass `ArgumnetParser` keyword arguments to `argsclass`.\nDescription is its own parameter - the rest are passed as the `parser_params` parameter as a `dict`.\n\nWhen a class is used as a subcommand (see below), `parser_params` are passed to `add_parser`, including `aliases`.\n```pycon\n>>> from datargs import parse, argsclass\n>>> @argsclass(description="Romans go home!", parser_params=dict(prog="messiah.py"))\n... class Args:\n...     flag: bool\n>>> parse(Args, ["-h"], parser=parser)\nusage: messiah.py [-h] [--flag]\nRomans go home!\n...\n```\n\nor you can pass your own parser:\n```pycon\n>>> from argparse import ArgumentParser\n>>> from datargs import parse, argsclass\n>>> @argsclass\n... class Args:\n...     flag: bool\n>>> parser = ArgumentParser(description="Romans go home!", prog="messiah.py")\n>>> parse(Args, ["-h"], parser=parser)\nusage: messiah.py [-h] [--flag]\nRomans go home!\n...\n```\n\nUse `make_parser()` to create a parser and save it for later:\n```pycon\n>>> from datargs import make_parser\n>>> @dataclass\n... class Args:\n...     ...\n>>> parser = make_parser(Args)  # pass `parser=...` to modify an existing parser\n```\n**NOTE**: passing your own parser ignores `ArgumentParser` params passed to `argsclass()`.\n\n### Enums\nWith `datargs`, enums Just Work™:\n\n```pycon\n>>> import enum, attr, datargs\n>>> class FoodEnum(enum.Enum):\n...     ham = 0\n...     spam = 1\n>>> @attr.dataclass\n... class Args:\n...     food: FoodEnum\n>>> datargs.parse(Args, ["--food", "ham"])\nArgs(food=<FoodEnum.ham: 0>)\n>>> datargs.parse(Args, ["--food", "eggs"])\nusage: enum_test.py [-h] --food {ham,spam}\nenum_test.py: error: argument --food: invalid choice: \'eggs\' (choose from [\'ham\', \'spam\'])\n```\n\n**NOTE**: enums are passed by name on the command line and not by value.\n\n## Sequences, Optionals, and Literals\nHave a `Sequence` or a `List` of something to\nautomatically use `nargs`:\n\n\n```python\nfrom pathlib import Path\nfrom dataclasses import dataclass\nfrom typing import Sequence\nfrom datargs import parse\n\n@dataclass\nclass Args:\n    # same as nargs=\'*\'\n    files: Sequence[Path] = ()\n\nargs = parse(Args, ["--files", "foo.txt", "bar.txt"])\nassert args.files == [Path("foo.txt"), Path("bar.txt")]\n```\n\nSpecify a list of positional parameters like so:\n\n```python\nfrom datargs import argsclass, arg\n@argsclass\nclass Args:\n    arg: Sequence[int] = arg(default=(), positional=True)\n```\n\n`Optional` arguments default to `None`:\n\n```python\nfrom pathlib import Path\nfrom dataclasses import dataclass\nfrom typing import Optional\nfrom datargs import parse\n\n@dataclass\nclass Args:\n    path: Optional[Path] = None\n\nargs = parse(Args, ["--path", "foo.txt"])\nassert args.path == Path("foo.txt")\n\nargs = parse(Args, [])\nassert args.path is None\n```\n\nAnd `Literal` can be used to specify choices:\n\n```python\nfrom pathlib import Path\nfrom dataclasses import dataclass\nfrom typing import Literal\nfrom datargs import parse\n\n@dataclass\nclass Args:\n    path: Literal[Path("foo.txt"), Path("bar.txt")]\n\nargs = parse(Args, ["--path", "foo.txt"])\nassert args.path == Path("foo.txt")\n\n# Throws an error!\nargs = parse(Args, ["--path", "bad-option.txt"])\n```\n\n### Sub Commands\n\nNo need to specify a useless `dest` to dispatch on different commands.\nA `Union` of dataclasses/attrs classes automatically becomes a group of subparsers.\nThe attribute holding the `Union` holds the appropriate instance\nupon parsing, making your code type-safe:\n\n```python\nimport typing, logging\nfrom datargs import argsclass, arg, parse\n\n@argsclass(description="install package")\nclass Install:\n    package: str = arg(positional=True, help="package to install")\n\n@argsclass(description="show all packages")\nclass Show:\n    verbose: bool = arg(help="show extra info")\n\n@argsclass(description="Pip Install Packages!")\nclass Pip:\n    action: typing.Union[Install, Show]\n    log: str = None\n\nargs = parse(Pip, ["--log", "debug.log", "install", "my_package"])\nprint(args)\n# prints: Pip(action=Install(package=\'my_package\'), log=\'debug.log\')\n\n# Consume arguments:\nif args.log:\n    logging.basicConfig(filename=args.log)\nif isinstance(args.action, Install):\n    install_package(args.action.package)\n    # static type error: args.action.verbose\nelif isinstance(args.action, Show):\n    list_all_packages(verbose=args.action.verbose)\nelse:\n    assert False, "Unreachable code"\n```\nCommand name is derived from class name. To change this, use the `name` parameter to `@argsclass`.\n\nAs with all other parameters to `add_parser`,\n`aliases` can be passed as a key in `parser_params` to add subcommand aliases.\n\n**NOTE**: if the commented-out line above does not issue a type error, try adding an `@dataclass/@attr.s`\nbefore or instead of `@argsclass()`:\n\n```python\n@argsclass(description="Pip Install Packages!")  # optional\n@dataclass\nclass Pip:\n    action: typing.Union[Install, Show]\n    log: str = None\n...\nif isinstance(args.action, Install):\n    install_package(args.action.package)\n    # this should now produce a type error: args.action.verbose\n```\n\n## "Why not"s and design choices\nMany libraries out there do similar things. This list serves as documentation for existing solutions and differences.\n\nSo, why not...\n\n### Just use argparse?\nThat\'s easy. The interface is clumsy and repetitive, a.k.a boilerplate. Additionally, `ArgumentParser.parse_args()` returns a `Namespace`, which is\nequivalent to `Any`, meaning that it any attribute access is legal when type checking. Alas, invalid attribute access will fail at runtime. For example:\n```python\ndef parse_args():\n    parser = ArgumentParser()\n    parser.add_argument("--url")\n    return parser.parse_args()\n\ndef main():\n    args = parse_args()\n    print(args.url)\n```\n\nLet\'s say for some reason `--url` is changed to `--uri`:\n\n```python\nparser.add_argument("--uri")\n...\nprint(args.url)  # oops\n```\nYou won\'t discover you made a mistake until you run the code. With `datargs`, a static type checker will issue an error.\nAlso, why use a carriage when you have a spaceship?\n\n### Use [`click`](https://click.palletsprojects.com/en/7.x/)?\n`click` is a great library. It provides many utilities for command line programs.\n\nUse `datargs` if you believe user interface should not be coupled with implementation, or if you\nwant to use `argparse` without boilerplate.\nUse `click` if you don\'t care.\n\n\n### Use [`clout`](https://clout.readthedocs.io/en/latest/index.html)?\nIt seems that `clout` aims to be an end-to-end solution for command line programs à la click.\n\nUse it if you need a broader solution. Use `datargs` if you want to use `argparse` without boilerplate.\n\n### Use [`simple-parsing`](https://pypi.org/project/simple-parsing/)?\nThis is another impressive library.\n\nUse it if you have deeply-nested options, or if the following points don\'t apply\nto you.\n\nUse `datargs` if you:\n* need `attrs` support\n* want as little magic as possible\n* don\'t have many options or they\'re not nested\n* prefer dashes (`--like-this`) over underscores (`--like_this`)\n\n### Use [`argparse-dataclass`](https://pypi.org/project/argparse-dataclass/)?\nIt\'s similar to this library. The main differences I found are:\n* no `attrs` support\n* not on github, so who you gonna call?\n\n### Use [`argparse-dataclasses`](https://pypi.org/project/argparse-dataclasses/)?\nSame points `argparse-dataclass` but also [Uses inheritance](https://refactoring.guru/replace-inheritance-with-delegation).\n\n## FAQs\n### Is this cross-platform?\nYes, just like `argparse`.\nIf you find a bug on a certain platform (or any other bug), please report it.\n\n### Why are mutually exclusive options not supported?\n\nThis library is based on the idea of a one-to-one correspondence between most parsers\nand simple classes. Conceptually, mutually exclusive options are analogous to\n[sum types](https://en.wikipedia.org/wiki/Tagged_union), just like [subparsers](#sub-commands) are,\nbut writing a class for each flag is not ergonomic enough.\nContact me if you want this feature or if you come up with a better solution.\n',
    'author': 'Roee Nizan',
    'author_email': 'roeen30@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roee30/datargs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
